﻿using System;
using System.Text;
using System.Runtime.Serialization;
using System.Xml.Serialization;
using System.Collections.Concurrent;
using System.Collections.Generic;
using System.Linq;

namespace OZX
{
    public static class OzCommonBytes
    {
        //*******************************************************
        // To Bytes
        //*******************************************************
        public static byte[] MessageToBytes(IOzxMessage msg)
        {
            int totalSize = 0;
            var segments = new List<byte[]>();

            append(segments, StringToBytes(msg.ID), ref totalSize);

            append(segments, StringToBytes(msg.FromConnection), ref totalSize);
            append(segments, StringToBytes(msg.FromAddress), ref totalSize);
            append(segments, StringToBytes(msg.FromStation), ref totalSize);

            append(segments, StringToBytes(msg.ToConnection), ref totalSize);
            append(segments, StringToBytes(msg.ToAddress), ref totalSize);
            append(segments, StringToBytes(msg.ToStation), ref totalSize);

            append(segments, StringToBytes(msg.Text), ref totalSize);
            
            append(segments, DateToBytes(msg.CreateDate), ref totalSize);
            append(segments, DateToBytes(msg.ValidUntil), ref totalSize);
            append(segments, DateToBytes(msg.TimeToSend), ref totalSize);
            
            append(segments, BoolToBytes(msg.IsSubmitReportRequested), ref totalSize);
            append(segments, BoolToBytes(msg.IsDeliveryReportRequested), ref totalSize);
            append(segments, BoolToBytes(msg.IsViewReportRequested), ref totalSize);

            append(segments, TagsToBytes(msg.GetTags()), ref totalSize);

            int offset = 0;
            var ret = new byte[totalSize];
            foreach (var b in segments)
            {
                Buffer.BlockCopy(b, 0, ret, offset, b.Length);
                offset += b.Length;
            }
            return ret;
        }

        static void append(List<byte[]> list, byte[] ret, ref int offset)
        {
            list.Add(ret);
            offset += ret.Length;
        }

        //*******************************************************
        // From Bytes
        //*******************************************************
        public static IOzxMessage BytesToMessage(byte[] bytes, ref int offset)
        {
            var msg = OzxMessageFactory.Instance.Create();

            msg.ID = BytesToString(bytes, ref offset);

            msg.FromConnection = BytesToString(bytes, ref offset);
            msg.FromAddress = BytesToString(bytes, ref offset);
            msg.FromStation = BytesToString(bytes, ref offset);

            msg.ToConnection = BytesToString(bytes, ref offset);
            msg.ToAddress = BytesToString(bytes, ref offset);
            msg.ToStation = BytesToString(bytes, ref offset);

            msg.Text = BytesToString(bytes, ref offset);

            msg.CreateDate = BytesToDate(bytes, ref offset);
            msg.ValidUntil = BytesToDate(bytes, ref offset);
            msg.TimeToSend = BytesToDate(bytes, ref offset);

            msg.IsSubmitReportRequested = BytesToBool(bytes, ref offset);
            msg.IsDeliveryReportRequested = BytesToBool(bytes, ref offset);
            msg.IsViewReportRequested = BytesToBool(bytes, ref offset);

            var tags = BytesToTags(bytes, ref offset);
            foreach (var tag in tags)
            {
                msg.AddTag(tag.Key, tag.Value);
            }
          
            return msg;
        }

        //*******************************************************
        // String encoding
        //*******************************************************

        public static byte[] StringToBytes(string s)
        {
            if (s == null) return new byte[4];
            var databytes = Encoding.UTF8.GetBytes(s);
            var lenbytes = BitConverter.GetBytes(databytes.Length);
            var bytes = new byte[lenbytes.Length + databytes.Length];
            Buffer.BlockCopy(lenbytes, 0, bytes, 0, lenbytes.Length);
            Buffer.BlockCopy(databytes, 0, bytes, lenbytes.Length, databytes.Length);
            return bytes;
        }

        public static string BytesToString(byte[] src, ref int offset)
        {
            var len = BitConverter.ToInt32(src, offset);
            offset += 4;
            var bytes = new byte[len];
            Buffer.BlockCopy(src, offset, bytes, 0, len);
            offset += len;
            var str = Encoding.UTF8.GetString(bytes);
            return str;
        }

        //*******************************************************
        // Date time encoding
        //*******************************************************

        public static byte[] DateToBytes(DateTime d)
        {
            var ticks = d.ToUniversalTime().Ticks;
            var bytes =  BitConverter.GetBytes(ticks);
            return bytes;
        }

        public static DateTime BytesToDate(byte[] src, ref int offset)
        {
            var ticks = BitConverter.ToInt64(src, offset);
            offset += 8;
            DateTime myUtcDateTime = DateTime.FromBinary(ticks);
            var localTime = myUtcDateTime.ToLocalTime();
            return localTime;
        }

        //*******************************************************
        // Int encoding
        //*******************************************************

        public static byte[] IntToBytes(int b)
        {
            var bytes = BitConverter.GetBytes(b);
            return bytes;
        }

        public static int BytesToInt(byte[] src, ref int offset)
        {
            var i = BitConverter.ToInt32(src, offset);
            offset += 4;
            return i;
        }

        //*******************************************************
        // Bool encoding
        //*******************************************************

        public static byte[] BoolToBytes(bool b)
        {
            var ret = new byte[1];
            if (b) ret[0] = 1;
            else ret[0] = 0;
            return ret;
        }

        public static byte BoolToByte(bool b)
        {
            if (b) return (byte)1;
            return (byte)0;
        }

        public static bool BytesToBool(byte[] src, ref int offset)
        {
            bool ret = false;
            if (src[offset] == 1) ret = true;
            offset++;
            return ret;
        }

        //*******************************************************
        // Tag encoding
        //*******************************************************

        public static byte[] TagsToBytes(Dictionary<string,string> tags)
        {
            var tagCountLenInBytes = IntToBytes(tags.Count);
            int totalSize = 0;
            var list = new List<Byte[]>();
            foreach (var tag in tags)
            {
                var key = StringToBytes(tag.Key);
                var val = StringToBytes(tag.Value);
                var tagBytes = new byte[key.Length + val.Length];
                Buffer.BlockCopy(key, 0, tagBytes, 0, key.Length);
                Buffer.BlockCopy(val, 0, tagBytes, key.Length, val.Length);
                list.Add(tagBytes);
                totalSize += tagBytes.Length;
            }
            int offset = 0;
            var ret = new byte[tagCountLenInBytes.Length + totalSize];
            Buffer.BlockCopy(tagCountLenInBytes, 0, ret, 0, tagCountLenInBytes.Length);
            offset += tagCountLenInBytes.Length;
            foreach (var b in list)
            {
                Buffer.BlockCopy(b, 0, ret, offset, b.Length);
                offset += b.Length;
            }
            return ret;
        }

        public static Dictionary<string,string> BytesToTags(byte[] bytes, ref int offset)
        {
            var ret = new Dictionary<string, string>();
            var tagCount = BytesToInt(bytes, ref offset);
            var end = offset + tagCount;
            while (offset < end)
            {
                var tagname = BytesToString(bytes, ref offset);
                var tagvalue = BytesToString(bytes, ref offset);
                ret.Add(tagname, tagvalue);
            }
            return ret;
        }
    }
}